

#include "thread.h"



// ======================================================================= Mutex



Mutex::Mutex()
{
    #ifdef __unix__
        pthread_mutex_init(&mutex, NULL);
    #elif __windows__
        mutex = CreateMutex(NULL, FALSE, NULL);
    #endif
}

Mutex::~Mutex()
{
    #ifdef __unix__
        pthread_mutex_destroy(&mutex);
    #elif __windows__
        CloseHandle(mutex);
    #endif
}

void Mutex::Lock()
{
    #ifdef __unix__
        pthread_mutex_lock(&mutex);
    #elif __windows__
        WaitForSingleObject(mutex, INFINITE);
    #endif
}

void Mutex::Unlock()
{
    #ifdef __unix__
        pthread_mutex_unlock(&mutex);
    #elif __windows__
        ReleaseMutex(mutex);
    #endif
}

bool Mutex::TryLock()
{
    #ifdef __unix__
        return pthread_mutex_trylock(&mutex);
    #elif __windows__
        return WaitForSingleObject(mutex, 0) == 0;
    #endif
}




// =========================================================== ConditionVariable



ConditionVariable::ConditionVariable()
    : Mutex()
{
    #ifdef __unix__
        pthread_cond_init(&cond, NULL);
    #elif __windows__
        InitializeConditionVariable(&cond);
    #endif
}

ConditionVariable::~ConditionVariable()
{
    #ifdef __unix__
        pthread_cond_destroy(&cond);
    #elif __windows__
        /*
        On Windows, Condition Variables don't need to be destroyed explicitly, see
        http://stackoverflow.com/questions/28975958/why-does-windows-have-no-deleteconditionvariable-function-to-go-together-with
        */
    #endif
}

bool ConditionVariable::Wait()
{
    #ifdef __unix__
        return pthread_cond_wait(&cond,&mutex) == 0;
    #elif __windows__
        return SleepConditionVariableCS(&cond, &mutex, INFINITE) != 0;
    #endif
}

void ConditionVariable::SignalOne()
{
    Lock();
    #ifdef __unix__
        pthread_cond_signal(&cond);
    #elif __windows__
        WakeConditionVariable(&cond);
    #endif
    Unlock();
}

void ConditionVariable::SignalAll()
{
    Lock();
    #ifdef __unix__
        pthread_cond_broadcast(&cond);
    #elif __windows__
        WakeAllConditionVariable(&cond);
    #endif
    Unlock();
}



// ============================================================= Thread


Thread::Thread()
{
    Started = false;
}

Thread::~Thread()
{
    Terminate();
}



#ifdef __unix__
    void* ThreadWrapper(void* t) {
#elif __windows__
    __cdecl void ThreadWrapper(void* t) {
#endif
    Thread* thread = (Thread*)(t);
    if(thread == NULL)
        throw "ThreadWrapper: parameter is not a Thread object";

    thread->RunThread();

    #ifdef __unix__
        return NULL;
    #endif
}

void Thread::Start()
{
    #ifdef __unix__
        pthread_create(&thread, NULL, &ThreadWrapper, this);
        //pthread_detach(thread);
    #elif __windows__
        thread = _beginthread(ThreadWrapper, 0, this);
    #endif
}

void Thread::Join()
{
    if(Started)
    {
        #ifdef __unix__
            pthread_join(thread, NULL);
        #elif __windows__
            WaitForSingleObject(thread, INFINITE);
        #endif
    }
    Started = false;
}

void Thread::Terminate()
{
    if(Started)
    {
        #ifdef __unix__
            pthread_cancel(thread);
        #elif __windows__
            TerminateThread((void*)thread,0);
        #endif
    }
    Started = false;
}

void Thread::TestTermination()
{
    #ifdef __unix__
        pthread_testcancel();
    #elif __windows__
    #endif
}




void Thread::Lock()
{
    mutex.Lock();
}

void Thread::Unlock()
{
    mutex.Unlock();
}

bool Thread::TryLock()
{
    return mutex.TryLock();
}



